#!/usr/bin/env python
# coding:utf-8
import datetime
import functools
import os
import sys
import uuid
from io import BytesIO as StringIO
from os import environ as env
import functools
import envparse
import pymongo
import tornado.auth
import tornado.options
from PIL import Image
from PIL.ImageDraw import ImageDraw
from fs.osfs import OSFS
from tornado.gen import coroutine
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.options import options, define, parse_command_line
from tornado.web import Application, RequestHandler, url, RedirectHandler, authenticated
from tornado.web import MissingArgumentError, HTTPError

if env.get('LOCAL', 'false').lower() == 'true':
    envparse.env.read_envfile("secrets.env")

VERSION = sys.version_info

SECURE_COOKIE = True

if VERSION.major >= 3:
    # noinspection PyUnresolvedReferences,PyCompatibility
    from urllib.parse import urlparse, urlunparse

if VERSION.major == 2:
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urlparse import urlparse, urlunparse


class Config(dict):
    def __init__(self, *args, **kwargs):

        kwargs['debug'] = kwargs.get('debug', False)
        kwargs['autoreload'] = kwargs.get('autoreload', False)

        super(Config, self).__init__(*args, **kwargs)

        for k, v in kwargs.items():
            if k not in options.as_dict():
                define(name=k, default=v)

        if 'port' not in options.as_dict():
            define(name='port', default=-1)

        parse_command_line()

        for k in kwargs.keys():
            self[k] = options[k]


mongo_client = pymongo.MongoClient(os.environ['MONGO_CONNECTION_STRING'])

mongo_database = mongo_client.get_database(name=os.environ['MONGO_DATABASE'])

__UPLOADS__ = "uploads/"
__STATIC_UPLOADS__ = os.path.join(os.path.dirname(__file__), __UPLOADS__)

__STATIC__ = "static/"
__STATIC_FILES__ = os.path.join(os.path.dirname(__file__), __STATIC__)
__FAVICON__ = os.path.join(os.path.dirname(__file__), __STATIC__, "favicon.ico")


class SessionManager(object):
    __USERS__ = mongo_database.get_collection("users")
    __SESSIONS__ = mongo_database.get_collection("sessions")

    @staticmethod
    def ensure_session_expire_index(expire_after_seconds=3600):
        _expire_index_name = 'expire_created_on'
        try:
            SessionManager.__SESSIONS__.create_index(
                keys=[("created_on", pymongo.ASCENDING)],
                expireAfterSeconds=expire_after_seconds,
                background=True,
                name=_expire_index_name)
        except pymongo.errors.OperationFailure:
            pass

    def __init__(self):
        super(SessionManager, self).__init__()

    def open_session(self, user, request=None):
        session = uuid.uuid4()
        sessid = str(session)
        self._save_session(session=session, user=user, request=request)
        return sessid

    def _save_session(self, session, user, request=None):
        session_doc = dict(
            session_id=session,
            user_id=user['email'],
            user_agent=request.headers.get('user-agent'),
            created_on=datetime.datetime.utcnow(),
            host=request.headers.get('host'))
        self.__SESSIONS__.insert_one(document=session_doc)
        return

    def register(self, user):
        user['_id'] = user['email']
        self.__USERS__.update(
            spec=dict(_id=user['email']),
            document=user,
            upsert=True)
        return

    # noinspection PyUnusedLocal
    def user_info(self, session_id):
        if type(session_id) is bytes:
            session = uuid.UUID(session_id.decode())
        else:  # type(session_id) is str:
            session = uuid.UUID(session_id)
        session_info = self.__SESSIONS__.find_one(dict(session_id=session)) or dict()
        user_id = session_info.get('user_id', None)
        user = self.__USERS__.find_one(dict(_id=user_id))
        return user


# noinspection PyAbstractClass
class BonsaiBaseRequestHandler(RequestHandler):
    __SESSION_COOKIE__ = 'sessid'

    session_manager = SessionManager()

    def get_current_user(self):
        if SECURE_COOKIE:
            session_id = self.get_secure_cookie(name=self.__SESSION_COOKIE__)
        else:
            session_id = self.get_cookie(name=self.__SESSION_COOKIE__)

        if session_id is None:
            return dict()

        user_info = self.session_manager.user_info(session_id=session_id)

        admins = ['serefguneysu@gmail.com']
        if user_info is None: return dict()

        if not user_info.get('email', None) in admins:
            self.redirect(url=self.reverse_url(name="unauthorized"))
            return False
        return user_info if user_info is not None else dict()


# noinspection PyAbstractClass
class GoogleOAuth2LoginHandler(
    BonsaiBaseRequestHandler,
    tornado.web.RequestHandler,
    tornado.auth.GoogleOAuth2Mixin):
    @tornado.gen.coroutine
    def get(self):
        google_auth_redirect2 = urlunparse((self.request.protocol,
                                            self.request.host,
                                            '/auth/google',
                                            None, None, None))

        if self.get_argument('code', False):

            user = yield self.get_authenticated_user(
                redirect_uri=google_auth_redirect2,
                code=self.get_argument('code'))

            # Save the user with e.g. set_secure_cookie
            access_token = user['access_token']

            user_info = yield self.oauth2_request(
                url='https://www.googleapis.com/oauth2/v2/userinfo',
                access_token=access_token)

            self.session_manager.register(user=user_info)
            session_id = self.session_manager.open_session(user=user_info, request=self.request)

            if SECURE_COOKIE:
                self.set_secure_cookie(name=self.__SESSION_COOKIE__, value=session_id, expires_days=1)
            else:
                self.set_cookie(name=self.__SESSION_COOKIE__, value=session_id, expires_days=1)

            self.clear_cookie(name='next')

            self.redirect(url=self.get_cookie(name='next'))


        else:
            self.set_cookie(name='next', value=self.get_argument(name='next'))
            yield self.authorize_redirect(
                redirect_uri=google_auth_redirect2,
                client_id=self.settings['google_oauth']['key'],
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})

    # noinspection PyMethodMayBeStatic
    def _on_auth(self, user):
        logging.debug(user)


class ImageTransformer(object):
    @staticmethod
    @functools.lru_cache(maxsize=100)
    def image_stream(filename, as_format):
        im = BaseImageHandler.image(filename)
        fp = StringIO()
        im.save(fp=fp, format=as_format)
        stream = fp.getvalue()
        return stream

    @staticmethod
    def image2stream(image, as_format):
        fp = StringIO()
        image.save(fp, format=as_format)
        return fp.getvalue()

    @staticmethod
    def thumbnail(image, size=(128, 128)):
        image.thumbnail(size)
        return image

    @staticmethod
    def image(path):
        im = Image.open((os.path.join(__UPLOADS__, path)))
        return im


# noinspection PyAbstractClass
class BaseImageHandler(RequestHandler, ImageTransformer):
    def set_image_header(self, fmt):
        self.set_header("Content-type", "image/%s" % fmt)

    def decide_image_format(self):
        fmt = self.get_argument("format", None)
        if fmt is not None:
            return fmt

        fmt = "webp" if self.request.headers.get('Accept', "*/*").find('image/webp') > -1 \
            else self.get_argument("format", "jpeg")
        return fmt


# noinspection PyAbstractClass
class SquareThumbnailHandler2(BaseImageHandler):
    size = (128, 128)

    @coroutine
    def get(self, filename, *args, **kwargs):
        as_format = self.decide_image_format()

        try:
            im = self.image(filename)
            # image = image.crop((40, 0, image.width - 10, image.height - 1)) # soldan 40px, sağdan 10px
            black_canvas = Image.new(mode="RGB", size=(max(im.size), max(im.size)), color=1)

            draw = ImageDraw(black_canvas)
            draw.rectangle(xy=black_canvas.getbbox(), fill="turquoise")

            paste_bbox = (
                int(black_canvas.width / 2) - int(im.width / 2),
                int(black_canvas.height / 2) - int(im.height / 2))
            black_canvas.paste(im, box=paste_bbox)
            black_canvas = black_canvas.resize(size=(256, 256))
            del im
            del draw

        except FileNotFoundError:
            raise HTTPError(status_code=404)

        # image = image.transpose(Image.FLIP_TOP_BOTTOM)
        image.thumbnail(self.size)
        # minh = min(image.getbbox()[2:])
        # image = image.crop(box=(0, 0, minh, minh))
        # print(image.getbbox()) # (x0, y0, x1, y1)
        self.set_image_header(as_format)
        self.finish(self.image2stream(black_canvas, as_format))


# noinspection PyAbstractClass
class SquareThumbnailHandler(BaseImageHandler):
    size = (128, 128)

    @coroutine
    def get(self, filename, *args, **kwargs):
        as_format = self.decide_image_format()

        try:
            im = self.image(filename)
            im = im.crop((40, 0, im.width - 10, im.height - 1))  # soldan 40px, sağdan 10px
            scaling_factor = 128 / max(im.size)
            im = im.resize(size=tuple(map(lambda x: int(scaling_factor * x), im.size)))
            black_canvas = Image.new(mode="RGB", size=(max(im.size), max(im.size)), color=1)

            draw = ImageDraw(black_canvas)
            draw.rectangle(xy=black_canvas.getbbox(), fill="black")

            paste_bbox = (
                int(black_canvas.width / 2) - int(im.width / 2),
                int(black_canvas.height / 2) - int(im.height / 2))
            black_canvas.paste(im, box=paste_bbox)
            del im
            del draw

        except FileNotFoundError:
            raise HTTPError(status_code=404)
        # image = image.transpose(Image.FLIP_TOP_BOTTOM)
        # image.thumbnail(self.size)
        # minh = min(image.getbbox()[2:])
        # image = image.crop(box=(0, 0, minh, minh))
        # print(image.getbbox()) # (x0, y0, x1, y1)
        self.set_image_header(as_format)
        self.finish(self.image2stream(black_canvas, as_format))


# noinspection PyAbstractClass,SpellCheckingInspection
class DynamicImageHandler(BaseImageHandler):
    def get(self, filename, *args, **kwargs):
        as_format = self.decide_image_format()
        try:
            stream = self.image_stream(filename, as_format)
        except FileNotFoundError as e:
            raise HTTPError(status_code=404)

        self.set_image_header(as_format)
        self.finish(stream)


# noinspection PyAbstractClass,SpellCheckingInspection
class FileUploadHelper:
    FS = OSFS(__UPLOADS__)

    @staticmethod
    def unique_filename(filename):
        return str(uuid.uuid4()) + os.path.splitext(filename)[1]

    def upload_async(self, filepart, folder):
        filename = self.unique_filename(filepart['filename'])
        FS = OSFS(root_path=os.path.join(__UPLOADS__), create=True)
        FS.setfile(path=filename, file=StringIO(filepart['body']))
        return filename


# noinspection PyAbstractClass
class UploadHandler(BonsaiBaseRequestHandler, RequestHandler, FileUploadHelper):
    @authenticated
    def post(self):
        try:
            func = functools.partial(self.upload_async, folder=self.current_user['email'])
            files = map(func, self.request.files['file'])
            self.set_cookie(name="uploaded_files", value=str.join(";", files))
            self.redirect(self.reverse_url(name="upload_successful"), status=303)

        except KeyError as e:
            raise MissingArgumentError(e.args[0])
        except Exception:
            raise

    @authenticated
    def get(self):
        self.render("upload.html", user=self.current_user)


# noinspection PyAbstractClass
class UploadSuccessfulHandler(BonsaiBaseRequestHandler, RequestHandler):
    def get(self, *args, **kwargs):
        files = self.get_cookie(name="uploaded_files").split(";")
        self.render("list.html", files=files, user=self.current_user)


# noinspection PyAbstractClass
class UploadFailedHandler(RequestHandler):
    def get(self, *args, **kwargs):
        error = self.get_cookie(name="upload_error")
        self.render("upload_failed.html", error=error)


# noinspection PyAbstractClass
class ProxyHandler(RequestHandler):
    def get(self, *args, **kwargs):
        self.finish(dict(args=args, kwargs=kwargs))


# noinspection PyAbstractClass
class FileListHandler(RequestHandler):
    def get(self, *args, **kwargs):
        files = os.listdir(os.path.join(__STATIC_UPLOADS__))
        self.render(template_name="list.html", files=files)


urls = [
    url(pattern=r'/upload', handler=UploadHandler, name='upload'),
    url(pattern=r"/upload_successful", handler=UploadSuccessfulHandler, name="upload_successful"),
    url(pattern=r"/upload_failed", handler=UploadFailedHandler, name="upload_failed"),
    url(pattern=r'/originals/'
                r'(?P<filename>[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}.\w{3,4})'
                r':square-thumbnail',
        handler=SquareThumbnailHandler, name='square-thumbnail'),
    url(pattern=r'/originals/'
                r'(?P<filename>[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}.\w{3,4})',
        handler=DynamicImageHandler, name='original'),
    url(pattern=r'/originals/', handler=FileListHandler, name="originals_list"),
    url(pattern=r'/originals', handler=RedirectHandler, kwargs=dict(url="/originals/")),
    url(pattern='/auth/google', handler=GoogleOAuth2LoginHandler, name='oauth.google'),
    # url(pattern=r'/originals/(.*)', handler=StaticFileHandler, kwargs={'path': __STATIC_UPLOADS__}, name="originals"),
    # url(pattern=r'/(.*)', handler=ProxyHandler, name='proxy'),
]

google_oauth = dict(
    key=env['OAUTH_GOOGLE_KEY'],
    secret=env['OAUTH_GOOGLE_SECRET']
)

settings = Config(
    logging='info',
    xsrf_cookies=True,
    debug=True,
    login_url='/auth/google',
    google_oauth=google_oauth,
    template_path=os.path.join(os.path.dirname(__file__), "templates"),
    static_path=os.path.join(os.path.dirname(__file__), "static"),
    cookie_secret=env['COOKIE_SECRET'],
)

application = Application(handlers=urls, **settings)


def main():
    SessionManager.ensure_session_expire_index(expire_after_seconds=24 * 60 * 60)

    server = HTTPServer(application)
    server.bind(8000)
    server.start(num_processes=1)
    IOLoop.instance().start()


if __name__ == '__main__':
    main()
